/*
 * Demonstration program for mini-X graphics.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nano-X.h"

static	GR_WINDOW_ID	w1;		/* id for large window */
static	GR_GC_ID	gc1;		/* graphics context for text */

void errorcatcher();			/* routine to handle errors */

int
main(int argc,char **argv)
{
	GR_EVENT	event;		/* current event */

#if !defined(HAVE_JPEG_SUPPORT) && !defined(HAVE_BMP_SUPPORT)
	printf("Sorry, no image support compiled in\n");
	exit(1);	
#endif

	if(argc != 2)
	{
		printf("Usage: demo3 <.bmp ");
#ifdef HAVE_JPEG_SUPPORT
		printf("or .jpg image>\n");
#else
		printf("image>\n");
#endif
		exit(1);
	}

	if (GrOpen() < 0) {
		fprintf(stderr, "cannot open graphics\n");
		exit(1);
	}
	
	GrSetErrorHandler(errorcatcher);

	w1 = GrNewWindow(GR_ROOT_WINDOW_ID, 10, 10, 400, 400, 4, BLACK, WHITE);

	GrMapWindow(w1);

	gc1 = GrNewGC();

	GrSetGCForeground(gc1, WHITE);

	if (strstr(argv[1], ".bmp") != NULL)
		GrBMP(w1, gc1, 0, 0, IMAGE_MAX_SIZE, IMAGE_MAX_SIZE, argv[1]);
#ifdef HAVE_JPEG_SUPPORT
	else
		GrJPEG(w1, gc1, 0, 0, IMAGE_MAX_SIZE, IMAGE_MAX_SIZE, FALSE, argv[1]);
#endif

	while (1) {GrCheckNextEvent(&event);}

	return 0;
}



/*
 * Here on an unrecoverable error.
 */
void
errorcatcher(code, name, id)
	GR_ERROR	code;		/* error code */
	GR_FUNC_NAME	name;		/* function name which failed */
	GR_ID		id;		/* resource id */
{
	GrClose();
	fprintf(stderr, "DEMO ERROR: code %d, function %s, resource id %d\n",
		code, name, id);
	exit(1);
}
